import re
from django.db import models
from django.utils.timezone import localtime
from dateutil import tz


DEVICE_STATUS_OFFLINE = "offline"
DEVICE_STATUS_ONLINE = "online"


def remove_spaces(inp: str) -> str:
    return re.sub(r"\s+", " ", inp)


def quote_str(inp: str) -> str:
    return inp.replace("\n", "\\n") \
        .replace("\r", "\\r") \
        .replace("\t", "\\t")


def elided(text: str, max_length: int) -> str:
        if len(text) < max_length:
            return text

        return f"{ text[0:max_length-3] }..."


class Device(models.Model):
    __tablename__ = "devices"

    class Meta :
        verbose_name_plural = "devices"

    name = models.CharField(max_length=100, unique=True)
    host = models.CharField(max_length=100)
    port = models.IntegerField(default=1001)
    enabled = models.BooleanField(default=True)
    allowed_to_update = models.BooleanField(default=True)
    status = models.CharField(max_length=100, default=DEVICE_STATUS_OFFLINE)

    streams = models.ManyToManyField("Stream", through="DeviceStreams")

    def __str__(self) -> str:
        return f"{ self.name } ({ self.host }:{ self.port })"


class Pos(models.Model):
    __tablename__ = "poses"

    class Meta :
        verbose_name_plural = "poses"

    device = models.ForeignKey(Device, on_delete=models.CASCADE)
    text = models.TextField()
    received_at = models.DateTimeField(auto_now_add=True)

    def __str__(self) -> str:
        return elided(f"[ { self.device.name } ] [ { localtime(self.received_at, timezone=tz.gettz('America/Sao_Paulo')) } ] - { remove_spaces(quote_str(self.text)) }", 150)


class MediaServer(models.Model):
    __tablename__ = "mediaservers"

    name = models.CharField(max_length=256)
    host = models.CharField(max_length=100)
    port = models.IntegerField(default=7001)

    username = models.CharField(max_length=512)
    password = models.CharField(max_length=512)

    def __str__(self) -> str:
        return f"{ self.name }({ self.host }:{ self.port })"


class Stream(models.Model):
    __tablename__ = "streams"

    server = models.ForeignKey(MediaServer, on_delete=models.CASCADE)
    name = models.CharField(max_length=256)
    nx_id = models.CharField(max_length=512)

    def __str__(self) -> str:
        return f"{ self.name }({ self.nx_id })"


class DeviceStreams(models.Model):
    __tablename__ = "device_streams"

    class Meta :
        verbose_name_plural = "device streams"

    device = models.ForeignKey(Device, on_delete=models.CASCADE)
    stream = models.ForeignKey(Stream, on_delete=models.CASCADE, related_name="streams")

    def __str__(self) -> str:
        return f"{ self.device } -> { self.stream }"


class Log(models.Model):
    __tablename__ = "device_streams"

    level = models.CharField(max_length=120, default="MSG")
    scope = models.CharField(max_length=120, default="Global")
    message = models.TextField()
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self) -> str:
        return f"[{ self.created_at }][{ self.level }] [{ self.scope }] - { self.message }"