import re
import sys
import asyncio
import datetime
from typing import Dict, List, Set
from dataclasses import dataclass, field
from functools import partial

import zmq
import zmq.asyncio
import aiohttp



@dataclass
class TaskInfo:
    name: str
    host: str
    port: int
    task: asyncio.Task

    stream_ids: List[str] = field(default_factory=lambda: [])
    server_host: str = ""
    server_user: str = ""
    server_pass: str = ""



task_by_name: Dict[str, TaskInfo] = dict()
task_by_name_lock: asyncio.Lock = asyncio.Lock()
device_names: Set[str] = set()



LOG_DIR = "/tmp/"
if sys.platform == 'win32':
    LOG_DIR = ""
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())



def log(msg: str):
    print(msg)
    with open(f"{ LOG_DIR }zclient.log", "a") as fp:
        fp.write(f"[{ datetime.datetime.now() }]{ msg }\n")



def debug(msg: str):
    log(f"DEBUG: { msg }")



def info(msg: str):
    log(f"INFO: { msg }")



def warn(msg: str):
    log(f"WARN: { msg }")



def error(msg: str):
    log(f"ERROR: { msg }")



async def process_data(dev_name: str, data: bytes, force: bool) -> bytes:
    debug(f"process_data(force: { force })")
    parts: List[bytes] = data.split(b"\n\n")
    debug(f"Num of parts: { len(parts) }")

    if (not force) and (len(parts)==1):
        return data

    for i, p in enumerate(parts):
        debug(f"Pos number: { i }")
        if (not force) and (i==len(parts)-1):
            break

        pstr = p.decode().strip()
        if len(pstr) == 0:
            continue

        info(f"POS:\n{ pstr }")

        async with aiohttp.ClientSession() as session:
            data = {
                "pos": pstr,
                "device": dev_name
            }
            async with session.post("http://127.0.0.1/create_pos", json=data) as response:
                pass

            # Send event
            debug("Sending event")
            task_info = task_by_name[dev_name]

            pstr = re.sub("\n", "_", pstr)
            now = datetime.datetime.now()
            url = f"""http://{ task_info.server_host }/api/createEvent?timestamp={ int(datetime.datetime.timestamp(now)) * 1000 }&source={ dev_name }&caption=POS_{ dev_name }&description={ pstr }&metadata={{"cameraRefs":[{ ','.join([ f'"{ id }"' for id in task_info.stream_ids ]) }]}}"""
            debug(url)
            async with session.get(url, auth=aiohttp.BasicAuth(task_info.server_user, task_info.server_pass)) as response:
                pass

    if force:
        return b""

    return parts[-1]



async def connect_to_device(name: str, host: str, port: int):
    reconnect = True
    all_data: bytes = b""

    while True:
        if reconnect:
            info("Reset all_data to empty")
            all_data = b""

            try:
                reader, writer = await asyncio.open_connection(host, port)
            except (asyncio.CancelledError, KeyboardInterrupt):
                raise
            except:
                error(f"Can't connect to { name } ({ host }:{ port }), retrying after 3 seconds...")
                await asyncio.sleep(3)
                continue

            info(f"Connected to { name } ({ host }:{ port })")
            reconnect = False

        try:
            data = await asyncio.wait_for(reader.read(100), timeout=5)
            if not data:
                if len(all_data) > 0:
                    await process_data(name, all_data, True)

                error(f"POS device { name } down, retrying after 3 seconds...")
                await asyncio.sleep(3)
                reconnect = True
                continue

            info(f"Just received:\n{ data.decode() }")

            all_data += data
            all_data = re.sub(b"\r\n", b"\n", all_data)

            all_data = await process_data(name, all_data, False)
        except ConnectionResetError:
            warn(f"Connection reset '{ name }', reconnect in 3 seconds")

            async with aiohttp.ClientSession() as session:
                async with session.post(f"http://127.0.0.1/offline/{ name }") as response:
                    pass

            try:
                writer.close()
            except:
                pass

            if len(all_data) > 0:
                await process_data(name, all_data, True)

            reconnect = True
            await asyncio.sleep(3)
            continue
        except (asyncio.CancelledError, KeyboardInterrupt):
            writer.close()

            if len(all_data) > 0:
                await process_data(name, all_data, True)

            raise
        except asyncio.TimeoutError:
            writer.write(b"\0")
            await writer.drain()
            continue



def task_finished(name: str, t: asyncio.Task):
    info(f"Task of '{ name }' finished")

    try:
        t.result()
    except asyncio.CancelledError:
        pass
    except Exception as e:
        warn(f"Exception: { e }")



async def create_task(name: str, host: str, port: int):
    global device_names
    global task_by_name
    global task_by_name_lock

    server_host = ""
    server_username = ""
    server_password = ""
    stream_ids = []

    async with aiohttp.ClientSession() as session:
        async with session.get(f"http://127.0.0.1/device/{ name }") as response:
            ret = await response.json()
            assert "servers" in ret
            assert "stream_ids" in ret

            servers = ret["servers"]
            for server in servers:
                assert "host" in server
                assert "password" in server
                assert "user" in server

                server_host = server["host"]
                server_password = server["password"]
                server_username = server["user"]

            stream_ids = ret["stream_ids"]

    task = asyncio.create_task(connect_to_device(name, host, port))
    task_info: TaskInfo = TaskInfo(
        name=name,
        host=host,
        port=port,
        task=task,
        server_host=server_host,
        server_user=server_username,
        server_pass=server_password,
        stream_ids=stream_ids)

    task.add_done_callback(partial(task_finished, name))

    async with task_by_name_lock:
        task_by_name[name] = task_info
        device_names.add(name)



async def ipc_task():
    ctx = zmq.asyncio.Context()
    socket = ctx.socket(zmq.SUB)
    socket.connect("tcp://127.0.0.1:6666")
    socket.subscribe("")

    while True:
        msg = await socket.recv_json()
        assert "name" in msg
        assert "host" in msg
        assert "port" in msg

        name = msg["name"]
        host = msg["host"]
        port = msg["port"]

        if name in task_by_name:
            ti = task_by_name[name]
            ti.task.cancel()
            try:
                await ti.task
            except:
                pass

            async with task_by_name_lock:
                del task_by_name[name]
                device_names.remove(name)

        await create_task(name, host, port)



async def sync_task():
    global device_names
    global task_by_name
    global task_by_name_lock

    while True:
        async with aiohttp.ClientSession() as session:
            async with session.get("http://127.0.0.1/devices") as response:
                devices = await response.json()
                dev_names: Set[str] = set()
                for dev in devices:
                    name = dev["name"]
                    host = dev["host"]
                    port = dev["port"]

                    dev_names.add(name)

                    if name in device_names:
                        continue

                    await create_task(name, host, port)

        need_to_stop = device_names.difference(dev_names)
        for name in need_to_stop:
            if name not in task_by_name:
                continue

            debug(f"Going to stop { name }")

            ti = task_by_name[name]
            ti.task.cancel()
            try:
                await ti.task
            except:
                pass

            async with task_by_name_lock:
                del task_by_name[name]
                device_names.remove(name)

        await asyncio.sleep(5)


async def main():
    await asyncio.gather(
        ipc_task(),
        sync_task())



asyncio.run(main())